/*
** Module   : Registry.java
** Abstract : Progress 4GL compatible Winodws registry related methods
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130604 Created initial version.
** 002 EVL 20140117 The native library call is performing now in dedicated place and only once.
**                  See the ClientCore class.
** 003 EVL 20140327 The default constructor needs the session parameter to get properties from
**                  server side.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.net.*;
import com.goldencode.util.*;

/**
 * A class implementing the Windows registry based environment methods.  Key values are stored in
 * registry under appropriate keys in registry subsection.  This is the default approach if none
 * new environment is not loaded and used by LOAD and USE Progress statements.
 * <p>
 * Some terms need to be clarified.  Unlike handling the INI file the 4GL sections are stored as
 * registry subkeys, while the keys are stored as registry value names.  Current investigation
 * show the 4GL base option has no effect in registry handling, everything is happening under
 * HKEY_CURRENT_USER registry base.  Also the documented behavior for getting the subkeys in
 * format 'section1@key1,section2@key2,..,sectionN@keyN' does not work.  Another format for
 * getting the list of the keys is implemented: 'section1,section2,..,sectionN'. 
 *
 * @author    EVL
 */
public class Registry
implements EnvironmentReader
{
   /** Windows registry key-subkey names separator. */
   public static final String NAME_SEPARATOR = "\\";
   
   /** Windows default hive name. */
   public static final String HKEY_DEFAULT_BASE = "HKEY_CURRENT_USER";
   
   /** Windows default key name. */
   public static final String KEY_DEFAULT = "";
   
   /**
    * Native entry point to get key value from registry.
    * 
    * @param base 
    *        The name of the root for the environment location.  It is currently ignored because
    *        the only root is working is HKEY_CURRENT_USER. 
    * @param section 
    *        The name of the section in the current environment registry entry to look the key.
    * @param key 
    *        The name of the key(or value name in a terms of the windows registry) whose value
    *        is requested.
    * 
    * @return The string representing the value of the given environment key.
    */
   private native static String getKeyVal(String base, String section, String key);

   /**
    * Getting the comma separated list of subkeys for the given environment name.
    * 
    * @param base 
    *        The name of the root for the environment location.  It is currently ignored because
    *        the only root is working is HKEY_CURRENT_USER. 
    * @param name 
    *        The name of the current environment registry entry to get the section list.
    * 
    * @return The string representing the comma separated list of the sections for a given
    *         environment entry.
    */
   private native static String getSubKeys(String base, String name);
   
   /**
    * Getting the comma separated list of keys(value name sin a terms of the windows registry)
    * for given subkey section.
    * 
    * @param base 
    *        The name of the root for the environment location.  It is currently ignored because
    *        the only root is working is HKEY_CURRENT_USER. 
    * @param section 
    *        The name of the section in the current environment registry entry to get the value
    *        names list.
    * 
    * @return The string representing the comma separated list of the keys in a given section.
    */
   private native static String getValueNames(String base, String section);
   
   /**
    * Native entry point to store new key value to registry.
    * 
    * @param base 
    *        The name of the root for the environment location.  It is currently ignored because
    *        the only root is working is HKEY_CURRENT_USER. 
    * @param section 
    *        The name of the section in the current environment registry entry to change the key.
    * @param key 
    *        The name of the key(or value name in a terms of the windows registry) to modify.
    * @param value 
    *        The new value to be stored for the key.
    * 
    * @return <code>true</code> in case of success, <code>false</code> otherwise.
    */
   private native static boolean setKeyVal(String base, String section, String key, String value);

   /**
    * Deletes name and it's value from subkey 'section'.
    * 
    * @param base 
    *        The name of the root for the environment location.  It is currently ignored because
    *        the only root is working is HKEY_CURRENT_USER. 
    * @param section 
    *        The name of the section in the current environment registry entry to remove the key.
    * @param key 
    *        The name of the key(or value name in a terms of the windows registry) to remove.
    * 
    * @return <code>true</code> in case of success, <code>false</code> otherwise.
    */
   private native static boolean removeValueName(String base, String section, String key);
   
   /**
    * Deletes subkey section and all content.
    * 
    * @param base 
    *        The name of the root for the environment location.  It is currently ignored because
    *        the only root is working is HKEY_CURRENT_USER. 
    * @param section 
    *        The name of the section to remove from the current environment registry entry.
    * 
    * @return <code>true</code> in case of success, <code>false</code> otherwise.
    */
   private native static boolean removeSubKey(String base, String section);

   /** The complete key path with subkeys to store the key values without base key name. */
   private String registryKeyName = null;
   
   /** The name of the registry base to store the keys. */
   private String baseKeyName = null;

   /** The flag indicating whether we need to create new registry sections replacing existed. */   
   private boolean isNew;

   /**
    * Default constructor creating the internals based on the expected registry keys for currently
    * installed Progress simulation.
    *
    * @param    session
    *           The server session to get the properties.
    */
   public Registry(Session session)
   {
      ServerPropertiesInspector spInsp =
         (ServerPropertiesInspector) RemoteObject.obtainNetworkInstance(
                                            ServerPropertiesInspector.class, session);
         
      registryKeyName = "SOFTWARE"+NAME_SEPARATOR+"PSC"+NAME_SEPARATOR+"PROGRESS"+NAME_SEPARATOR+
                        spInsp.getVersion();
      isNew = false;
      baseKeyName = HKEY_DEFAULT_BASE;
   }

   /**
    * Construct an instance of the environment with Windows registry as backend.
    *
    * @param env 
    *        The environment to create or load. Can be the registry key.
    * @param flagNew
    *        If <code>true</code> - the new file will be created overwriting the possibly
    *        existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under the particular base
    *        key value.
    */
   public Registry(String env, boolean flagNew, String baseKey)
   {
      registryKeyName = new String(env);
      isNew = flagNew; 
      if (baseKey != null)
      {
         baseKeyName = new String(baseKey);
      }
      else
      {
         baseKeyName = HKEY_DEFAULT_BASE;
      }
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    */
   public void load()
   {
      return;
   }

   /**
    * Adds, modifies and deletes keys or sections in current environment.  The real work it
    * performs is controlled by the input parameters.  In a simplest case when all three inputs
    * have the real not null and not empty string the method change the registry key for the
    * current environment.
    * <p>
    * The other possible cases are("" means the parameter is null or empty string if the
    * difference is not important):
    * section == "", key != "", value != "" causes adding directly under current environment
    *                                       value name 'key' with given 'value'.
    * section == "", key != "", value == "" removing value name 'key' directly under current
    *                                       environment with its 'value'
    * section != "", key != "", value == "" deletes name and it's value from subkey 'section'.
    * section != "", key == "", value == "" deletes subkey 'section' with all content.
    * section != "", key == null, value != "" sets new value for DEFAULT value name for subkey
    *                                         'section'.
    * section != "", key == null, value == "" removes DEFAULT value name setting 'value' as unset
    *                                         for subkey 'section'
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public void setKeyValue(String section, String key, String value)
   {
      // Prepare initial parameters for JNI call
      String sectionInitial = null;
      String keyInitial = null;
      String valueInitial = null;

      if (section == null || section.isEmpty()) // Section is empty or null
      {
         if (key != null && !key.isEmpty()) // Key is not empty
         {
            if (value != null && !value.isEmpty()) // Value is not empty
            {
               // Adding name-value directly under current environment
               sectionInitial = registryKeyName;
               keyInitial = key;
               valueInitial = value;
               // Call C native code here
               setKeyVal(baseKeyName, sectionInitial, keyInitial, valueInitial);
            }
            else  // Value is empty or null
            {
               // Deletes name and it's value from current environment (no subsections)
               sectionInitial = registryKeyName;
               keyInitial = key;
               // Call C native code here
               removeValueName(baseKeyName, sectionInitial, keyInitial);
            }
         }
         else // Key is empty or null
         {
            // Other 2 invalid cases - probably we have to generate error here
         }
      }
      else
      {
         if (key == null) // Key is null, working with DEFAULT key for given section
         {
            if (value == null || value.isEmpty()) // Value is empty or null
            {
               // Deletes subkey section and all content
               sectionInitial = registryKeyName + NAME_SEPARATOR + section;
               keyInitial = KEY_DEFAULT;
               // Call C native code here
               removeValueName(baseKeyName, sectionInitial, keyInitial);
            }
            else // Value is not empty
            {
               // Sets the DEFAULT value name for subkey section
               sectionInitial = registryKeyName + NAME_SEPARATOR + section;
               keyInitial = KEY_DEFAULT;
               valueInitial = value;
               // Call C native code here
               setKeyVal(baseKeyName, sectionInitial, keyInitial, valueInitial);
            }
         }
         else if (key.isEmpty()) // Key is empty
         {
            if (value == null || value.isEmpty()) // Value is empty or null
            {
               // Deletes subkey section and all content
               sectionInitial = registryKeyName + NAME_SEPARATOR + section;
               // Call C native code here
               removeSubKey(baseKeyName, sectionInitial);
            }
         }
         else // Key is not empty
         {
            if (value == null || value.isEmpty()) // Value is empty
            {
               // Deletes name and it's value from subkey 'section'
               sectionInitial = registryKeyName + NAME_SEPARATOR + section;
               keyInitial = key;
               // Call C native code here
               removeValueName(baseKeyName, sectionInitial, keyInitial);
            }
            else // Value is not empty
            {
               // Usual work, just set the new value
               sectionInitial = registryKeyName + NAME_SEPARATOR + section;
               keyInitial = key;
               valueInitial = value;
               // Call C native code here
               setKeyVal(baseKeyName, sectionInitial, keyInitial, valueInitial);
            }
         }
      }
   }

   /**
    * Getting the key value or key or section lists from the current environment.  The real work
    * it performs is controlled by the input parameters.  In a simplest case when all two inputs
    * have the real not null and not empty string the method gets the value of the given key.
    * <p>
    * The other possible cases are("" means the parameter is null or empty string if the
    * difference is not important):
    * section == "", key != "" gets the value for given value name 'key' under current
    *                          environment.  There is no 'section' for this key.
    * section == "", key == "" gets the comma separated list of all subkeys for the current
    *                          environment.
    * section != "", key == empty gets the comma separated list of value names for subkey
    *                             'section'.  Empty here means not null string with 0 length.
    * section != "", key == null gets the DEFAULT value for the given subkey 'section'.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public String getKeyValue(String section, String key)
   {
      // Prepare initial parameters for JNI call
      String sectionInitial = null;
      String keyInitial = null;
      String valueInitial = null;
      // What we return from C native code 
      String result = null;

      if (section == null || section.isEmpty()) // Section is empty or null
      {
         if (key == null || key.isEmpty()) // Key is empty or null
         {
            // Getting the comma separated list of subkey@value_name pair
            sectionInitial = registryKeyName;
            result = getSubKeys(baseKeyName, sectionInitial);
         }
         else // Key is not empty
         {
            // Getting the value of the key under current environment
            sectionInitial = registryKeyName;
            keyInitial = key;
            // Call C native code here
            result = getKeyVal(baseKeyName, sectionInitial, keyInitial);
         }
      }
      else // Section is not empty
      {
         if (key == null) // Key is null
         {
            // Getting the default value name for given subkey section
            sectionInitial = registryKeyName + NAME_SEPARATOR + section;
            keyInitial = KEY_DEFAULT;
            // Call C native code here
            result = getKeyVal(baseKeyName, sectionInitial, keyInitial);
         }
         else if (key.isEmpty()) // Key is empty but not null, special case
         {
            // Getting the comma separated list of value_names for given subkey section
            sectionInitial = registryKeyName + NAME_SEPARATOR + section;
            result = getValueNames(baseKeyName, sectionInitial);
         }
         else // Section and key are not empty
         {
            // Usual provessing, get the key value 
            sectionInitial = registryKeyName + NAME_SEPARATOR + section;
            keyInitial = key;
            // Call C native code here
            result = getKeyVal(baseKeyName, sectionInitial, keyInitial);
         }
      }

      return result;
   }

   /**
    * Unloads specifies environment from the current one.
    */
   public void unload()
   {
      return;
   }
}
